from arjuna.setu.types import SetuManagedObject
from arjuna.setuext.guiauto.impl.element.guielement import GuiElement
from arjuna.tpi.enums import ArjunaOption

class BasicWindow(SetuManagedObject):

    def __init__(self, automator):
        super().__init__()
        self.__automator = automator
        self.__window_handle = None
        self.__config = automator.config

    @property
    def config(self):
        return self.__config    

    @property
    def automator(self):
        return self.__automator

    @property
    def handle(self):
        return self.__window_handle

    def _set_handle(self, handle):
        self.__window_handle = handle

    def focus(self):
        self.automator.dispatcher.focus_on_window(self.handle)

    def is_main_window(self):
        return False

    def set_window_size(self, width, height):
        self.automator.dispatcher.set_current_window_size(width, height)

    def maximize(self):
        self.automator.dispatcher.maximize_current_window()

    def get_title(self):
        return self.automator.dispatcher.get_current_window_title()

    def get_size(self):
        return self.automator.dispatcher.get_current_window_size()

class MainWindow(BasicWindow):

    def __init__(self, automator):
        super().__init__(automator)
        self._set_handle(self.get_current_window_handle()) 
        self.__all_child_windows = {}
        self.__setu_id_map = {}
        self.__resize_window_as_per_config()

    def get_all_child_window_handles(self):
        handles = self.automator.dispatcher.get_all_window_handles()
        handles = [handle for handle in handles if handle != self.handle]
        new_handles = []
        for handle in handles:
            if handle != self.handle:
                if handle not in self.__all_child_windows:
                    cwin = ChildWindow(self.automator, self, handle)
                    self.__all_child_windows[handle] = cwin
                    self.__setu_id_map[cwin.setu_id] = cwin
                    new_handles.append(handle)
        return handles, new_handles

    def get_latest_child_window(self):
        _, new_handles = self.get_all_child_window_handles()
        if not new_handles:
            raise Exception("No new window was launched.")
        elif len(new_handles) > 1:
            raise Exception("Multiple new windows were launched, so can not deterministically jump to a new window.")
        return self.__all_child_windows[new_handles[0]]

    def __get_child_window(self, handle):
        return self.__all_child_windows[handle]

    def __focus_on_window(self, handle):
        self.__get_child_window(handle).jump()

    def delete_window(self, setu_id, handle):
        del self.__all_child_windows[handle]
        del self.__setu_id_map[setu_id]

    def close_all_child_windows(self):
        all_child_handles, _ = self.get_all_child_window_handles()
        for handle in all_child_handles:
            cwin = self.__get_child_window(handle)
            cwin.focus()
            cwin.close()
        self.focus()

    def get_current_window_handle(self):
        return self.automator.dispatcher.get_current_window_handle()

    def get_window_for_setu_id(self, setu_id):
        if self.setu_id == setu_id:
            return self
        else:
            return self.__setu_id_map[setu_id]

    def get_window_for_locator(self, locator_type, locator_value):
        all_child_handles, _ = self.get_all_child_window_handles()
        for handle in all_child_handles:
            cwin = self.__get_child_window(handle)
            cwin.focus()
            if locator_type.lower() == "window_title":
                if cwin.get_title() == locator_value:
                    return cwin
                else:
                    try:
                        element = self.automator.create_element_with_locator(locator_type, locator_value)
                        element.find()
                        return cwin
                    except:
                        continue
        raise Exception("No child window contains an element with locator type: {} and locator value: {}".format(locator_type, locator_value))

    def __resize_window_as_per_config(self):
        # Resize window
        config = self.config
        browser_width = config.setu_config.value(ArjunaOption.BROWSER_DIM_WIDTH)
        browser_height = config.setu_config.value(ArjunaOption.BROWSER_DIM_HEIGHT)
        should_maximize = config.setu_config.value(ArjunaOption.BROWSER_MAXIMIZE)

        if config.setu_config.is_not_set(ArjunaOption.BROWSER_DIM_WIDTH) and config.setu_config.is_not_set(ArjunaOption.BROWSER_DIM_HEIGHT):
            if should_maximize:
                self.maximize()
        else:
            width, height = None, None
            current_width, current_height = self.get_size()
            width = config.setu_config.is_not_set(ArjunaOption.BROWSER_DIM_WIDTH) and browser_width or current_width
            height = config.setu_config.is_not_set(ArjunaOption.BROWSER_DIM_HEIGHT) and browser_height or current_height
            self.set_window_size(width, height)

    def is_main_window(self):
        return True

    def close(self):
        raise Exception("You can not close main window. Use automator.quit() to quit application.")

class ChildWindow(BasicWindow):

    def __init__(self, automator, main_window, handle):
        super().__init__(automator)
        self.__main_window = main_window
        self._set_handle(handle) 

    def close(self):
        self.focus()
        self.automator.dispatcher.close_current_window()
        self.__main_window.delete_window(self.setu_id, self.handle)
        self.__main_window.focus()


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import threading
from collections import OrderedDict
from arjuna.tpi import Arjuna
from arjuna.lib.thread.decorators import *
from arjuna.unitee.reporter.result.types import SteppedResult

class MinimalConsoleReporter:

    def __init__(self):
        self.lock = threading.RLock()
        self.console = Arjuna.get_console()

    @sync_method('lock')
    def __print(self, heading, rdict={}):
        self.console.marker(25)
        self.console.display(heading)
        self.console.marker(25)
        for k,v in rdict.items():
            if v != "-":
                self.console.display_multiline_key_value(k,v)

    def update_info(self, rdict):
        # Does not print informational messages
        pass

    def update_issue(self, issue):
        # Prints full issue information
        self.__print("Issue Notification", issue)

    def update_test_object_result(self, reportable):
        # Prints only test result
        rdict = reportable.result
        if rdict['otype'] != 'Test': return
        self.console.display("Result:: [{}:{}] {} || Module:{}.{} Function:{} Test#{} || Stage:{} Group:{}".format(
            rdict['rtype'],
            rdict['rcode'],
            (rdict['rtype'] in {'FAIL', 'ERROR'}) and ("Issue# " + str(rdict['iid'])) or "",
            rdict['pkg'],
            rdict['module'],
            rdict['function'],
            rdict['test'],
            rdict['stage'],
            rdict['group']
        ))

    def set_up(self):
        pass

    def tear_down(self):
        pass

    def update_info_headers(self, headers):
        pass

    def update_execution_headers(self, headers):
        pass

    def update_step_headers(self, headers):
        pass

    def update_issue_headers(self, headers):
        pass


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from arjuna.lib.exceptions import *

class DataSourceFinished(StopIteration):
    def __init__(self, msg=None):
        super().__init__(msg is None and "Done" or msg)


class EmptyListDataRecordLookupException(Exception):
    def __init__(self, index):
        super().__init__("Invalid index [%s] used for list data record lookup. It is empty.".format(index))


class ListDataRecordLookupException(Exception):
    def __init__(self, index, max_index):
        super().__init__(
            "Invalid index [%s] used for list data record lookup. Use indices between 0 and %d".format(index,
                                                                                                       max_index))


class MapDataRecordLookupException(Exception):
    def __init__(self, key):
        super().__init__("Invalid Key/header [%s] used for map data record lookup.".format(key))


class DataSourceConstructionError(Exception):
    def __init__(self, message, name, exc):
        super().__init__(message)
        self.name = name
        self.exc = exc

    def get_Exception(self):
        return self.exc

    def get_Name(self):
        return self.name


class InvalidTestObjectException(Exception):
    pass


class SessionNodesFinishedException(Exception):
    def __init__(self):
        super().__init__("Done")


class SubTestsFinished(Exception):
    def __init__(self):
        super().__init__("Done")


class TestGroupsFinishedException(Exception):
    def __init__(self):
        super().__init__("Done")


class PickerMisConfigurationException(Exception):
    def __init__(self):
        super().__init__("Picker is misconfigured.")

# Test Result Related

class StepResultEvent(ArjunaException):
    def __init__(self, step):
        super().__init__(step.assert_message)
        self.step = step

class Pass(StepResultEvent):
    def __init__(self, check):
        super().__init__(check)


class Error(StepResultEvent):
    def __init__(self, step):
        super().__init__(step)


class Failure(StepResultEvent):
    def __init__(self, step):
        super().__init__(step)

class DependencyNotMet(Exception):
    def __init__(self, iid):
        self.iid = iid


from arjuna.client.core.action import SetuActionType
from arjuna.client.core.connector import SetuArg
from .automator import AbstractAppAutomator

from arjuna.tpi.guiauto import With


class BaseGui(AbstractAppAutomator):

    def __init__(self, automator, label=None, def_file_name=None, parent=None, register=True):
        super().__init__(automator.get_config())

        self.__automator = automator
        self.__children_map = dict()
        self.__auto_context = automator.get_automation_context()
        self.__test_session = automator.get_test_session()
        self._set_automator_setu_id_arg(automator.get_setu_id())
        self.__gui_registered = False
        self.__parent = parent

        self.__label = None
        self.__def_file_name = None

        if label:
            self.set_label(label)

        if def_file_name:
            self.set_def_file_name(def_file_name)

        if register:
            self._register()

    def __check_reg_status(self):
        if self.__gui_registered:
            raise Exception("Attempt to change Gui critical attribute post registration with Setu.")

    def is_gui(self):
        return True

    def set_label(self, label):
        self.__check_reg_status()
        self.__label = label
        self.set_def_file_name(label + ".gns")

    def set_def_file_name(self, name):
        self.__check_reg_status()
        self.__def_file_name = name

    def _register(self):
        if self.__gui_registered:
            raise Exception("Attempt to re-register Gui with Setu.")

        args = [
            SetuArg.arg("testSessionSetuId", self.__test_session.get_setu_id()),
            SetuArg.arg("automatorSetuId", self.__automator.get_setu_id()),
            SetuArg.arg("label", self.__label),
            SetuArg.arg("name", self.__class__.__name__),
            SetuArg.arg("qualName", self.__class__.__qualname__),
            SetuArg.arg("defFileName", self.__def_file_name)
        ]

        gui_setu_id = None
        if not self.__parent:
            gui_setu_id = self.__test_session.create_gui(self.__automator, *args)
        else:
            args.append(SetuArg.arg("parentGuiSetuId", self.__parent.get_setu_id()))
            response = self._send_request(SetuActionType.GUIAUTO_GUI_CREATE_GUI, *args)
            gui_setu_id = response.get_gui_setu_id()

        self._set_setu_id(gui_setu_id)
        self._set_self_setu_id_arg("guiSetuId")

        if self.__parent:
            self.__parent.add_child(self.__label, self)

        self.__load()

    def _get_automator(self):
        return self.__automator

    def get_qualified_name(self):
        return self.__class__.__qualname__

    def add_child(self, label, gui):
        self.__children_map[label.lower()] = gui

    def get_child(self, label):
        if label:
            if label.lower() in self.__children_map:
                return self.__children_map[label.lower()]
            else:
                raise Exception("No child Gui with label: {} defined.".format(label))
        else:
            raise Exception("Child Gui label is None.")

    def reach_until(self):
        # Children can override and write any necessary loading instructions
        pass

    def validate_readiness(self):
        pass

    def __load(self):
        try:
            self.validate_readiness()
        except:
            try:
                self.reach_until()
                self.validate_readiness()
            except Exception as e:
                raise Exception(
                    "UI [{}] with SetuId [{}] did not load as expected. Error: {}.",
                    self.__class__.__name__,
                    self.get_setu_id(),
                    str(e)
                )

    def __convert_with(self, input):
        if isinstance(input, With):
            return input
        else:
            return With.assigned_name(input)

    def Element(self, name_or_with_obj):
        return super().Element(self.__convert_with(name_or_with_obj))

    def MultiElement(self, name_or_with_obj):
        return super().MultiElement(self.__convert_with(name_or_with_obj))

    def DropDown(self, name_or_with_obj):
        return super().DropDown(self.__convert_with(name_or_with_obj))

    def RadioGroup(self, name_or_with_obj):
        return super().RadioGroup(self.__convert_with(name_or_with_obj))

    def Alert(self, name_or_with_obj):
        return super().Alert(self.__convert_with(name_or_with_obj))

    def Frame(self, name_or_with_obj):
        return super().Frame(self.__convert_with(name_or_with_obj))

    def ChildWindow(self, name_or_with_obj):
        return super().ChildWindow(self.__convert_with(name_or_with_obj))

    def MainWindow(self):
        return self.get_automator().MainWindow()

    def DomRoot(self):
        return self.get_automator().DomRoot()

    def Browser(self):
        return self.get_automator().Browser()

    def get_automator(self):
        return self.__automator


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import threading
import sys
import xlwt

from collections import OrderedDict
from arjuna.lib.thread.decorators import *
from arjuna.unitee.reporter.result.types import SteppedResult
from arjuna.unitee.enums import *

# The autofit solution has been taken from StackOverflow thread
# https://stackoverflow.com/questions/6929115/python-xlwt-accessing-existing-cell-content-auto-adjust-column-width
class _ExcelSheet:

    def __init__(self, wb, name):
        self.__wb = wb
        self.__name = name
        self.__current_row_counter = -1
        self.ws = self.__wb.add_sheet(self.__name)
        self.__header_style = xlwt.easyxf('font: bold on; borders: left thin, right thin, top thin, bottom thin; align: vertical top;\
                                          pattern: pattern solid, fore_colour coral;')
        self.__row_style = xlwt.easyxf('borders: left thin, right thin, top thin, bottom thin; align: vertical top')
        self.__row_style.alignment.wrap = 1
        self.__widths = dict()

    def __write_row(self, data, style):
        import arial10
        self.__current_row_counter += 1
        for index, item in enumerate(data):
            item = str(item)
            self.ws.write(self.__current_row_counter, index, item, style)
            width = int(arial10.fitwidth(item))
            if width > self.__widths.get(index,0):
                self.__widths[index] = width
                self.ws.col(index).width = width

    def update_headers(self, headers):
        self.__write_row(headers, self.__header_style)

    def update_entry(self, record):
        self.__write_row(record, self.__row_style)

class ExcelReporter:

    def __init__(self):
        self.lock = threading.RLock()
        self.path = ArjunaCore.config.value(UniteePropertyEnum.RUN_REPORT_DIR) + "/ArjunaTestReport.xls"

    def set_up(self):
        self.wb = xlwt.Workbook()
        self.info_sheet = _ExcelSheet(self.wb, "Notifications")
        self.exec_sheet = _ExcelSheet(self.wb, "Execution Report")
        self.steps_sheet = _ExcelSheet(self.wb, "Steps")
        self.issues_sheet = _ExcelSheet(self.wb, "Issues")
        self._execheaders = None

    def tear_down(self):
        self.wb.save(self.path)

    def update_info_headers(self, headers):
        self.info_sheet.update_headers(headers)

    def update_execution_headers(self, headers):
        self.exec_sheet.update_headers(headers)
        self._execheaders = headers

    def update_step_headers(self, headers):
        self.steps_sheet.update_headers(list(self._execheaders)[:10] + list(headers))

    def update_issue_headers(self, headers):
        self.issues_sheet.update_headers(headers)

    @sync_method('lock')
    def _update(self, sheet, record):
        sheet.update_entry(record)

    def update_info(self, rdict):
        self._update(self.info_sheet, rdict.values())

    def update_issue(self, issue):
        self._update(self.issues_sheet, issue.values())

    @sync_method('lock')
    def update_test_object_result(self, reportable):
        self._update(self.exec_sheet, reportable.result.values())
        if reportable.has_steps():
            first = True
            prefix = ["-" for i in range(10)]
            for step in reportable.steps:
                try:
                    if first:
                        self._update(self.steps_sheet, list(reportable.result.values())[:10] + list(step.values()))
                        first=False
                    else:
                        self._update(self.steps_sheet, prefix + list(step.values()))
                except Exception as e:
                    print ("Exception occured")
                    print (e)
                    import traceback
                    traceback.print_exc()
                    sys.exit(1)



from arjuna.setu.webclient.requester import SetuAgentRequester


class SetuActorRequester():

    def __init__(self, base_url):
        super().__init__()
        self.__agent_requester = SetuAgentRequester(base_url)

    def post(self, url_suffix, json_dict):
        return self.__agent_requester.post(url_suffix, json_dict)


'''
This file is a part of Test Mile Arjuna
Copyright 2018 Test Mile Software Testing Pvt Ltd

Website: www.TestMile.com
Email: support [at] testmile.com
Creator: Rahul Verma

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import os
import sys
import platform

def is_os_windows():
    return platform.system() == "Windows"

def cexit():
	from arjuna.tpi import Arjuna
	Arjuna.get_console().display_error("Exiting...")
	sys.exit()

def fexit():
	print("Exiting because of Fatal Error...", file=sys.stderr)
	import time
	time.sleep(0.5)
	sys.exit(1)

# public static Runtime get_run_time() {
#     return Runtime.get_runtime();
# }

# public static String getOSName() {
#     return System.get_property("os.name");
# }

def get_line_separator():
    return os.linesep


def get_path_separator():
    return os.sep



import platform
import os
from enum import Enum
from arjuna.tpi.enums import *

from arjuna.setu.types import SetuManagedObject
from arjuna.lib.types.descriptors import *


class Config(SetuManagedObject):
    DESKTOP_CONTEXTS = {GuiAutomationContext.NATIVE, GuiAutomationContext.WEB}
    MOBILE_WEB_CONTEXTS = {GuiAutomationContext.ANDROID_WEB, GuiAutomationContext.IOS_WEB}
    ALL_WEB_CONTEXTS = {GuiAutomationContext.WEB, GuiAutomationContext.ANDROID_WEB, GuiAutomationContext.IOS_WEB}
    MOBILE_NATIVE_CONTEXTS = {GuiAutomationContext.ANDROID_NATIVE, GuiAutomationContext.IOS_NATIVE}

    def __init__(self):
        super().__init__()
        self.__setu_config = None
        self.__user_config = None
        self.__processor = None

    @property
    def setu_config(self):
        return self.__setu_config

    @setu_config.setter
    def setu_config(self, conf):
        self.__setu_config = conf

    @property
    def user_config(self):
        return self.__user_config

    @user_config.setter
    def user_config(self, conf):
        self.__user_config = conf

    @property
    def processor(self):
        return self.__processor

    @processor.setter
    def processor(self, processor):
        self.__processor = processor

    def as_json_dict(self):
        return {
            "arjunaOptions": self.__setu_config.as_json_dict(),
            "userOptions": self.__user_config.as_json_dict()
        }

    def __modify_bin_name_for_windows(self, name):
        if platform.system().lower() == "windows":
            return name + ".exe"
        else:
            return name

    def __get_driver_path(self, name):
        existing_driver_path = self.setu_config.value(ArjunaOption.SELENIUM_DRIVER_PATH)
        not_set_yet_str = "<DRIVER_NAME>"
        if existing_driver_path.find(not_set_yet_str) != -1:
            return self.setu_config.value(ArjunaOption.SELENIUM_DRIVER_PATH).replace(not_set_yet_str, self.__modify_bin_name_for_windows(name))
        else:
            # Some other driver might have been set
            return os.path.join(os.path.dirname(existing_driver_path), name)

    def process_arjuna_options(self):
        for_browser = {
            BrowserName.CHROME: {
                ArjunaOption.SELENIUM_DRIVER_PROP : "webdriver.chrome.driver",
                ArjunaOption.SELENIUM_DRIVER_PATH : self.__get_driver_path("chromedriver")
            },

            BrowserName.FIREFOX: {
                ArjunaOption.SELENIUM_DRIVER_PROP : "webdriver.gecko.driver",
                ArjunaOption.SELENIUM_DRIVER_PATH : self.__get_driver_path("geckodriver")
            },

            BrowserName.SAFARI: {
                ArjunaOption.SELENIUM_DRIVER_PROP : "webdriver.safari.driver",
                ArjunaOption.SELENIUM_DRIVER_PATH : self.__get_driver_path("safaridriver")
            }
        }

        browser = self.setu_config.get_browser_name()
        for opt, opt_value in for_browser[browser].items():
            self.setu_config._config_dict[opt] = opt_value

    def update(self, override_config):
        self.setu_config.update(override_config.setu_config)
        self.user_config.update(override_config.user_config)


class AbstractConfig:

    def __init__(self, config_dict):
        self.__config_dict = config_dict

    @property
    def _config_dict(self):
        return self.__config_dict

    def is_not_set(self, key):
        self._validate_key(key)
        try:
            return self.value(key).upper() == "NOT_SET"
        except:
            return False

    def _validate_key(self, key):
        pass

    def __value(self, v):
        if isinstance(v, Enum):
            return v.name
        elif type(v) is list:
            out = []
            for value in v:
                if isinstance(value, Enum):
                    out.append(value.name)
                else:
                    out.append(value)
            return out
        else:
            return v

    def value(self, key):
        self._validate_key(key)
        v = self.__config_dict[key]
        return self.__value(v)

    def as_map(self):
        return self.__config_dict

    def __as_enum_name_or_same(self, input):
        if isinstance(input, Enum):
            return input.name
        else:
            return input

    def as_json_dict(self):
        return {k:self.__value(v) for k,v in self.as_map().items()}

    def update(self, override_config):
        self.__config_dict.update(override_config.as_map())

    def enumerate(self):
        from arjuna import ArjunaCore
        keys = list(self.central_config.arjuna_options.keys())
        keys.sort()
        ArjunaCore.console.marker(100)
        header = " Central Properties Table "
        mark_length = (50 - len(header)// 2)
        ArjunaCore.console.marker_on_same_line(mark_length)
        ArjunaCore.console.display_on_same_line(header)
        ArjunaCore.console.marker(mark_length)
        ArjunaCore.console.marker(100)
        for key in keys:
            if self.central_config.arjuna_options[key].visible:
                sval = self.central_config.arjuna_options[key].value
                if EnumConstant.check(sval):
                    sval = sval.name
                elif EnumConstantList.check(sval):
                    sval = str([i.name for i in sval])
                else:
                    if key != "arjuna.root.dir" and String.check(sval):
                        sval = sval.replace(ARJUNA_ROOT, "<arjuna_root_dir>")
                    else:
                        sval = str(sval)

                ArjunaCore.console.display(
                    "| {:60s}| {}".format(self.central_config.arjuna_options[key].text,
                                          sval))
                ArjunaCore.console.display("| {:60s}| {}".format("(" + key + ")", ""))
                ArjunaCore.console.marker(100)
        ArjunaCore.console.marker(100)


class UserConfig(AbstractConfig):

    def __init__(self, config_dict):
        super().__init__(config_dict)


class SetuConfig(AbstractConfig):

    def __init__(self, config_dict):
        super().__init__(config_dict)

    def _validate_key(self, key):
        if not isinstance(key, ArjunaOption):
            raise Exception("Key must be an enum consts of type ArjunaOption")

    def as_json_dict(self):
        out = {k.name: v for k,v in super().as_json_dict().items()}
        return out

    def get_gui_automator_name(self):
        return self._config_dict[ArjunaOption.AUTOMATOR_NAME]

    def get_guiauto_actor_mode(self):
        return self._config_dict[ArjunaOption.SETU_GUIAUTO_ACTOR_MODE]

    def get_guiauto_context(self):
        return self._config_dict[ArjunaOption.GUIAUTO_CONTEXT]

    def get_browser_name(self):
        return BrowserName[self.value(ArjunaOption.BROWSER_NAME)]

    def get_host_os(self):
        return DesktopOS[self.value(ArjunaOption.TESTRUN_HOST_OS)]

    def has_desktop_context(self):
        return self.get_guiauto_context() in Config.DESKTOP_CONTEXTS

    def has_mobile_web_context(self):
        return self.get_guiauto_context() in Config.MOBILE_WEB_CONTEXTS

    def has_mobile_native_context(self):
        return self.get_guiauto_context() in Config.MOBILE_NATIVE_CONTEXTS

    def has_web_context(self):
        return self.get_guiauto_context() in Config.ALL_WEB_CONTEXTS



import uuid
import abc
import types

class SetuManagedObject:

    def __init__(self):
        self.__setu_id = str(uuid.uuid4())

    @property
    def setu_id(self):
        return self.__setu_id

    def get_setu_id(self):
        return self.__setu_id

class SetuConfiguredObject(SetuManagedObject, metaclass=abc.ABCMeta):

    def __init__(self, config):
        super().__init__()
        self.__dispatcher_creator = None
        self.__dispatcher = None
        self.__config = config

    @property
    def config(self):
        return self.__config

    @property
    def dispatcher(self):
        return self.__dispatcher

    def _set_dispatcher(self, dispatcher):
        self.__dispatcher = dispatcher

    @property
    def dispatcher_creator(self):
        return self.__dispatcher_creator

    @dispatcher_creator.setter
    def dispatcher_creator(self, creator):
        self.__dispatcher_creator = creator
        self.create_dispatcher()

    @abc.abstractmethod
    def create_dispatcher(self):
        pass